@@ -9,10 +9,11 @@
 from .enums import (
     ENUM_D_TAG, ENUM_E_VERSION, ENUM_P_TYPE_BASE, ENUM_SH_TYPE_BASE,
     ENUM_RELOC_TYPE_i386, ENUM_RELOC_TYPE_x64,
-    ENUM_RELOC_TYPE_ARM, ENUM_RELOC_TYPE_AARCH64, ENUM_RELOC_TYPE_MIPS,
-    ENUM_ATTR_TAG_ARM)
-from .constants import P_FLAGS, SH_FLAGS, SUNW_SYMINFO_FLAGS, VER_FLAGS
-from ..common.py3compat import iteritems
+    ENUM_RELOC_TYPE_ARM, ENUM_RELOC_TYPE_AARCH64, ENUM_RELOC_TYPE_PPC64,
+    ENUM_RELOC_TYPE_MIPS, ENUM_ATTR_TAG_ARM, ENUM_DT_FLAGS, ENUM_DT_FLAGS_1)
+from .constants import (
+    P_FLAGS, RH_FLAGS, SH_FLAGS, SUNW_SYMINFO_FLAGS, VER_FLAGS)
+from ..common.py3compat import bytes2hex, iteritems
 
 
 def describe_ei_class(x):
@@ -34,7 +35,13 @@ def describe_ei_osabi(x):
     return _DESCR_EI_OSABI.get(x, _unknown)
 
 
-def describe_e_type(x):
+def describe_e_type(x, elffile=None):
+    if elffile is not None and x == 'ET_DYN':
+        # Detect whether this is a normal SO or a PIE executable
+        dynamic = elffile.get_section_by_name('.dynamic')
+        for t in dynamic.iter_tags('DT_FLAGS_1'):
+            if t.entry.d_val & ENUM_DT_FLAGS_1['DF_1_PIE']:
+                return 'DYN (Position-Independent Executable file)'
     return _DESCR_E_TYPE.get(x, _unknown)
 
 
@@ -62,12 +69,28 @@ def describe_p_flags(x):
     return s
 
 
+def describe_rh_flags(x):
+    return ' '.join(
+        _DESCR_RH_FLAGS[flag]
+        for flag in (RH_FLAGS.RHF_NONE, RH_FLAGS.RHF_QUICKSTART,
+                     RH_FLAGS.RHF_NOTPOT, RH_FLAGS.RHF_NO_LIBRARY_REPLACEMENT,
+                     RH_FLAGS.RHF_NO_MOVE, RH_FLAGS.RHF_SGI_ONLY,
+                     RH_FLAGS.RHF_GUARANTEE_INIT,
+                     RH_FLAGS.RHF_DELTA_C_PLUS_PLUS,
+                     RH_FLAGS.RHF_GUARANTEE_START_INIT, RH_FLAGS.RHF_PIXIE,
+                     RH_FLAGS.RHF_DEFAULT_DELAY_LOAD,
+                     RH_FLAGS.RHF_REQUICKSTART, RH_FLAGS.RHF_REQUICKSTARTED,
+                     RH_FLAGS.RHF_CORD, RH_FLAGS.RHF_NO_UNRES_UNDEF,
+                     RH_FLAGS.RHF_RLD_ORDER_SAFE)
+        if x & flag)
+
+
 def describe_sh_type(x):
     if x in _DESCR_SH_TYPE:
         return _DESCR_SH_TYPE.get(x)
     elif (x >= ENUM_SH_TYPE_BASE['SHT_LOOS'] and
           x < ENUM_SH_TYPE_BASE['SHT_GNU_versym']):
-        return 'loos+%lx' % (x - ENUM_SH_TYPE_BASE['SHT_LOOS'])
+        return 'loos+0x%lx' % (x - ENUM_SH_TYPE_BASE['SHT_LOOS'])
     else:
         return _unknown
 
@@ -78,8 +101,12 @@ def describe_sh_flags(x):
             SH_FLAGS.SHF_WRITE, SH_FLAGS.SHF_ALLOC, SH_FLAGS.SHF_EXECINSTR,
             SH_FLAGS.SHF_MERGE, SH_FLAGS.SHF_STRINGS, SH_FLAGS.SHF_INFO_LINK,
             SH_FLAGS.SHF_LINK_ORDER, SH_FLAGS.SHF_OS_NONCONFORMING,
-            SH_FLAGS.SHF_GROUP, SH_FLAGS.SHF_TLS, SH_FLAGS.SHF_EXCLUDE):
+            SH_FLAGS.SHF_GROUP, SH_FLAGS.SHF_TLS, SH_FLAGS.SHF_MASKOS,
+            SH_FLAGS.SHF_EXCLUDE):
         s += _DESCR_SH_FLAGS[flag] if (x & flag) else ''
+    if not x & SH_FLAGS.SHF_EXCLUDE:
+        if x & SH_FLAGS.SHF_MASKPROC:
+            s += 'p'
     return s
 
 
@@ -95,6 +122,17 @@ def describe_symbol_visibility(x):
     return _DESCR_ST_VISIBILITY.get(x, _unknown)
 
 
+def describe_symbol_local(x):
+    return '[<localentry>: ' + str(1 << x) + ']'
+
+
+def describe_symbol_other(x):
+    vis = describe_symbol_visibility(x['visibility'])
+    if x['local'] > 1 and x['local'] < 7:
+        return vis + ' ' + describe_symbol_local(x['local'])
+    return vis
+
+
 def describe_symbol_shndx(x):
     return _DESCR_ST_SHNDX.get(x, '%3s' % x)
 
@@ -109,6 +147,8 @@ def describe_reloc_type(x, elffile):
         return _DESCR_RELOC_TYPE_ARM.get(x, _unknown)
     elif arch == 'AArch64':
         return _DESCR_RELOC_TYPE_AARCH64.get(x, _unknown)
+    elif arch == '64-bit PowerPC':
+        return _DESCR_RELOC_TYPE_PPC64.get(x, _unknown)
     elif arch == 'MIPS':
         return _DESCR_RELOC_TYPE_MIPS.get(x, _unknown)
     else:
@@ -119,6 +159,16 @@ def describe_dyn_tag(x):
     return _DESCR_D_TAG.get(x, _unknown)
 
 
+def describe_dt_flags(x):
+    return ' '.join(key[3:] for key, val in
+        sorted(ENUM_DT_FLAGS.items(), key=lambda t: t[1]) if x & val)
+
+
+def describe_dt_flags_1(x):
+    return ' '.join(key[5:] for key, val in
+        sorted(ENUM_DT_FLAGS_1.items(), key=lambda t: t[1]) if x & val)
+
+
 def describe_syminfo_flags(x):
     return ''.join(_DESCR_SYMINFO_FLAGS[flag] for flag in (
         SUNW_SYMINFO_FLAGS.SYMINFO_FLG_CAP,
@@ -148,21 +198,30 @@ def describe_note(x):
     n_desc = x['n_desc']
     desc = ''
     if x['n_type'] == 'NT_GNU_ABI_TAG':
-        desc = '\n    OS: %s, ABI: %d.%d.%d' % (
-            _DESCR_NOTE_ABI_TAG_OS.get(n_desc['abi_os'], _unknown),
-            n_desc['abi_major'], n_desc['abi_minor'], n_desc['abi_tiny'])
+        if x['n_name'] == 'Android':
+            desc = '\n   description data: %s ' % bytes2hex(x['n_descdata'])
+        else:
+            desc = '\n    OS: %s, ABI: %d.%d.%d' % (
+                _DESCR_NOTE_ABI_TAG_OS.get(n_desc['abi_os'], _unknown),
+                n_desc['abi_major'], n_desc['abi_minor'], n_desc['abi_tiny'])
     elif x['n_type'] == 'NT_GNU_BUILD_ID':
         desc = '\n    Build ID: %s' % (n_desc)
+    elif x['n_type'] == 'NT_GNU_GOLD_VERSION':
+        desc = '\n    Version: %s' % (n_desc)
+    elif x['n_type'] == 'NT_GNU_PROPERTY_TYPE_0':
+        desc = '\n      Properties: ' + describe_note_gnu_properties(x['n_desc'])
     else:
-        desc = '\n    description data: {}'.format(' '.join(
-            '{:02x}'.format(ord(byte)) for byte in n_desc
-        ))
+        desc = '\n      description data: {}'.format(bytes2hex(n_desc))
 
-    note_type = (x['n_type'] if isinstance(x['n_type'], str)
-                 else 'Unknown note type:')
-    note_type_desc = ('0x%.8x' % x['n_type']
-                      if isinstance(x['n_type'], int) else
-                      _DESCR_NOTE_N_TYPE.get(x['n_type'], _unknown))
+    if x['n_type'] == 'NT_GNU_ABI_TAG' and x['n_name'] == 'Android':
+        note_type = 'NT_VERSION'
+        note_type_desc = 'version'
+    else:
+        note_type = (x['n_type'] if isinstance(x['n_type'], str)
+                    else 'Unknown note type:')
+        note_type_desc = ('0x%.8x' % x['n_type']
+                        if isinstance(x['n_type'], int) else
+                        _DESCR_NOTE_N_TYPE.get(x['n_type'], _unknown))
     return '%s (%s)%s' % (note_type, note_type_desc, desc)
 
 
@@ -184,7 +243,7 @@ def describe_attr_tag_arm(tag, val, extra):
 
         elif tag == 'TAG_NODEFAULTS':
             return _DESCR_ATTR_TAG_ARM[tag] + 'True'
-            
+
         s = _DESCR_ATTR_TAG_ARM[tag]
         s += '"%s"' % val if val else ''
         return s
@@ -193,6 +252,70 @@ def describe_attr_tag_arm(tag, val, extra):
         return _DESCR_ATTR_TAG_ARM[tag] + d_entry[val]
 
 
+def describe_note_gnu_property_x86_feature_1(value):
+    descs = []
+    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS:
+        if value & mask:
+            descs.append(desc)
+    return 'x86 feature: ' + ', '.join(descs)
+
+def describe_note_gnu_property_x86_feature_2_used(value):
+    descs = []
+    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS:
+        if value & mask:
+            descs.append(desc)
+    return 'x86 feature used: ' + ', '.join(descs)
+
+def describe_note_gnu_property_x86_isa_1(value, verb):
+    descs = []
+    for mask, desc in _DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS:
+        if value & mask:
+            descs.append(desc)
+    return 'x86 ISA %s: %s' % (verb, ', '.join(descs))
+
+def describe_note_gnu_properties(properties):
+    descriptions = []
+    for prop in properties:
+        t, d, sz = prop.pr_type, prop.pr_data, prop.pr_datasz
+        if t == 'GNU_PROPERTY_STACK_SIZE':
+            if type(d) is int:
+                prop_desc = 'stack size: 0x%x' % d
+            else:
+                prop_desc = 'stack size: <corrupt length: 0x%x>' % sz
+        elif t == 'GNU_PROPERTY_NO_COPY_ON_PROTECTED':
+            if sz != 0:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = 'no copy on protected'
+        elif t == 'GNU_PROPERTY_X86_FEATURE_1_AND':
+            if sz != 4:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = describe_note_gnu_property_x86_feature_1(d)
+        elif t == 'GNU_PROPERTY_X86_FEATURE_2_USED':
+            if sz != 4:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = describe_note_gnu_property_x86_feature_2_used(d)                
+        elif t == 'GNU_PROPERTY_X86_ISA_1_NEEDED':
+            if sz != 4:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = describe_note_gnu_property_x86_isa_1(d, "needed")
+        elif t == 'GNU_PROPERTY_X86_ISA_1_USED':
+            if sz != 4:
+                prop_desc = ' <corrupt length: 0x%x>' % sz
+            else:
+                prop_desc = describe_note_gnu_property_x86_isa_1(d, "used")
+        elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOPROC <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIPROC:
+            prop_desc = '<processor-specific type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
+        elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOUSER <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIUSER:
+            prop_desc = '<application-specific type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
+        else:
+            prop_desc = '<unknown type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
+        descriptions.append(prop_desc)
+    return '\n        '.join(descriptions)
+
 #-------------------------------------------------------------------------------
 _unknown = '<unknown>'
 
@@ -232,6 +355,7 @@ _DESCR_EI_OSABI = dict(
     ELFOSABI_SORTIX='Sortix',
     ELFOSABI_ARM_AEABI='ARM - EABI',
     ELFOSABI_ARM='ARM - ABI',
+    ELFOSABI_CELL_LV2='CellOS Lv-2',
     ELFOSABI_STANDALONE='Standalone App',
 )
 
@@ -264,6 +388,7 @@ _DESCR_E_MACHINE = dict(
     EM_AARCH64='AArch64',
     EM_BLACKFIN='Analog Devices Blackfin',
     EM_PPC='PowerPC',
+    EM_PPC64='PowerPC64',
     RESERVED='RESERVED',
 )
 
@@ -279,6 +404,7 @@ _DESCR_P_TYPE = dict(
     PT_GNU_EH_FRAME='GNU_EH_FRAME',
     PT_GNU_STACK='GNU_STACK',
     PT_GNU_RELRO='GNU_RELRO',
+    PT_GNU_PROPERTY='GNU_PROPERTY',
     PT_ARM_ARCHEXT='ARM_ARCHEXT',
     PT_ARM_EXIDX='EXIDX',  # binutils calls this EXIDX, not ARM_EXIDX
     PT_AARCH64_ARCHEXT='AARCH64_ARCHEXT',
@@ -315,6 +441,7 @@ _DESCR_SH_TYPE = dict(
     SHT_GNU_HASH='GNU_HASH',
     SHT_GROUP='GROUP',
     SHT_SYMTAB_SHNDX='SYMTAB SECTION INDICIES',
+    SHT_RELR='RELR',
     SHT_GNU_verdef='VERDEF',
     SHT_GNU_verneed='VERNEED',
     SHT_GNU_versym='VERSYM',
@@ -358,6 +485,7 @@ _DESCR_SH_TYPE = dict(
     SHT_MIPS_EH_REGION='MIPS_EH_REGION',
     SHT_MIPS_XLATE_OLD='MIPS_XLATE_OLD',
     SHT_MIPS_PDR_EXCEPTION='MIPS_PDR_EXCEPTION',
+    SHT_MIPS_ABIFLAGS='MIPS_ABIFLAGS',
 )
 
 
@@ -372,10 +500,31 @@ _DESCR_SH_FLAGS = {
     SH_FLAGS.SHF_OS_NONCONFORMING: 'O',
     SH_FLAGS.SHF_GROUP: 'G',
     SH_FLAGS.SHF_TLS: 'T',
+    SH_FLAGS.SHF_MASKOS: 'o',
     SH_FLAGS.SHF_EXCLUDE: 'E',
 }
 
 
+_DESCR_RH_FLAGS = {
+    RH_FLAGS.RHF_NONE: 'NONE',
+    RH_FLAGS.RHF_QUICKSTART: 'QUICKSTART',
+    RH_FLAGS.RHF_NOTPOT: 'NOTPOT',
+    RH_FLAGS.RHF_NO_LIBRARY_REPLACEMENT: 'NO_LIBRARY_REPLACEMENT',
+    RH_FLAGS.RHF_NO_MOVE: 'NO_MOVE',
+    RH_FLAGS.RHF_SGI_ONLY: 'SGI_ONLY',
+    RH_FLAGS.RHF_GUARANTEE_INIT: 'GUARANTEE_INIT',
+    RH_FLAGS.RHF_DELTA_C_PLUS_PLUS: 'DELTA_C_PLUS_PLUS',
+    RH_FLAGS.RHF_GUARANTEE_START_INIT: 'GUARANTEE_START_INIT',
+    RH_FLAGS.RHF_PIXIE: 'PIXIE',
+    RH_FLAGS.RHF_DEFAULT_DELAY_LOAD: 'DEFAULT_DELAY_LOAD',
+    RH_FLAGS.RHF_REQUICKSTART: 'REQUICKSTART',
+    RH_FLAGS.RHF_REQUICKSTARTED: 'REQUICKSTARTED',
+    RH_FLAGS.RHF_CORD: 'CORD',
+    RH_FLAGS.RHF_NO_UNRES_UNDEF: 'NO_UNRES_UNDEF',
+    RH_FLAGS.RHF_RLD_ORDER_SAFE: 'RLD_ORDER_SAFE',
+}
+
+
 _DESCR_ST_INFO_TYPE = dict(
     STT_NOTYPE='NOTYPE',
     STT_OBJECT='OBJECT',
@@ -451,6 +600,7 @@ _DESCR_NOTE_N_TYPE = dict(
     NT_GNU_HWCAP='DSO-supplied software HWCAP info',
     NT_GNU_BUILD_ID='unique build ID bitstring',
     NT_GNU_GOLD_VERSION='gold version',
+    NT_GNU_PROPERTY_TYPE_0='program properties'
 )
 
 
@@ -464,6 +614,42 @@ _DESCR_NOTE_ABI_TAG_OS = dict(
     ELF_NOTE_OS_SYLLABLE='Syllable',
 )
 
+
+# Values in GNU .note.gnu.property notes (n_type=='NT_GNU_PROPERTY_TYPE_0') have
+# different formats which need to be parsed/described differently
+_DESCR_NOTE_GNU_PROPERTY_TYPE_LOPROC=0xc0000000
+_DESCR_NOTE_GNU_PROPERTY_TYPE_HIPROC=0xdfffffff
+_DESCR_NOTE_GNU_PROPERTY_TYPE_LOUSER=0xe0000000
+_DESCR_NOTE_GNU_PROPERTY_TYPE_HIUSER=0xffffffff
+
+
+# Bit masks for GNU_PROPERTY_X86_FEATURE_1_xxx flags in the form
+# (mask, flag_description) in the desired output order
+_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS = (
+    (1, 'IBT'),
+    (2, 'SHSTK'),
+    (4, 'LAM_U48'),
+    (8, 'LAM_U57'),
+)
+
+# Bit masks for GNU_PROPERTY_X86_FEATURE_2_xxx flags in the form
+# (mask, flag_description) in the desired output order
+_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS = (
+    (1, 'x86'),
+    (2, 'x87'),
+    (4, 'MMX'),
+    (8, 'XMM'),
+    (16, 'YMM'),
+    (32, 'ZMM'),
+)
+
+# Same for GNU_PROPERTY_X86_SET_1_xxx
+_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS = (
+    (1, 'x86-64-baseline'),
+    # TODO; there is a long list
+)
+
+
 def _reverse_dict(d, low_priority=()):
     """
     This is a tiny helper function to "reverse" the keys/values of a dictionary
@@ -484,6 +670,7 @@ _DESCR_RELOC_TYPE_i386 = _reverse_dict(ENUM_RELOC_TYPE_i386)
 _DESCR_RELOC_TYPE_x64 = _reverse_dict(ENUM_RELOC_TYPE_x64)
 _DESCR_RELOC_TYPE_ARM = _reverse_dict(ENUM_RELOC_TYPE_ARM)
 _DESCR_RELOC_TYPE_AARCH64 = _reverse_dict(ENUM_RELOC_TYPE_AARCH64)
+_DESCR_RELOC_TYPE_PPC64 = _reverse_dict(ENUM_RELOC_TYPE_PPC64)
 _DESCR_RELOC_TYPE_MIPS = _reverse_dict(ENUM_RELOC_TYPE_MIPS)
 
 _low_priority_D_TAG = (
